-- proasic3/PA3_genlib.vhdl
-- created 20190618
-- version 20190809 (restart)
-- version sam. août 10 08:06:58 CEST 2019
-- version dim. août 11 17:36:51 CEST 2019
-- version lun. août 12 18:36:04 CEST 2019
-- version mar. août 13 23:14:14 CEST 2019
-- version mar. nov. 19 07:04:27 CET 2019 : off-by-one in histogram display
-- version jeu. nov. 21 03:50:23 CET 2019 : changing from fast/trace to run_mode,
--    lookitupX_fast merged,  Gate_Census now with no parameter list
-- version dim. déc.  1 14:00:01 CET 2019 : preparing for 64 bits for histograms and changes
--
-- Released under the GNU AGPLv3 license
--
-- This file is the root library that is used by gategen.vhdl,
-- PA3_components.vhdl and PA3_definitions.vhdl.

Library ieee;
    use ieee.std_logic_1164.all;
use STD.textio.all;

package PA3_genlib is

  -- set to 61 for recent GHDL or VHDL'2018
  -- otherwise, set to 31 for standard '93 but limited capacity.
--  constant how_many_bits : integer := 61;

  -- this one uses less memory as well, OK for small or medium tests.
  constant how_many_bits : integer := 31;

  type big_uint_t is range                 0 to 2**(how_many_bits+1) - 1;
  type big_sint_t is range -2**how_many_bits to 2** how_many_bits    - 1;

  constant empty_str : string := "";
  -- that's useful in some places, where WRITE doesn't like concatenations...

  subtype LookupType1  is std_logic_vector(0 to  1);
  subtype LookupType2  is std_logic_vector(0 to  3);
  subtype LookupType4  is std_logic_vector(0 to  7);
  subtype LookupType8  is std_logic_vector(0 to 15);

  type enum_run_modes is (
    fast, trace, probe
  );
  shared variable run_mode : enum_run_modes := fast;

  -- the storage area for the histograms:
  -- !!!!!!!!!! should be better coded !!!!!!!!!!! (but works well so far)
  constant SPECIAL : natural := 16; -- the index of the counter of "non-binary" codes
  type type_Histo8 is array (0 to SPECIAL) of big_sint_t; -- can store a LUT8, the "shadow/meta" and the "special"s counter
  type type_Histo_entry is
    record
      index   : natural;     -- the index of the last lookup
      last    : natural;     -- (size of the LUT) the last valid index (can be 0)
      histo   : type_Histo8; -- the histogram
      curOut,                -- the last result of lookup
      prevOut : std_logic;   -- the previous result of the lookup
      changes : big_uint_t;  -- how many times the output changed
    end record;
  type type_Histo_Array is array(integer range <>) of type_Histo_entry;
  type type_Histo_Access is access type_Histo_Array;
  shared variable HistoPtr : type_Histo_Access;
  shared variable gate_instance_counter : integer := 0;

  -- the list of registered gates:
  -- (in case the general list of gates contains holes,
  -- such as GND/VCC (LUT0) gates that are not listed)
  type integer_array is array(integer range <>) of integer;
  type integer_array_ptr is access integer_array;
  shared variable indexPtr : integer_array_ptr;
  shared variable registered_instance_counter : integer := 0;

  shared variable gate_select_number : integer := 0;
  shared variable bit_select_number : integer := -1;
  shared variable cycle_counter : natural := 0;
  shared variable flip_LUT : integer := 0;

  type string3x2 is array(0 to 1) of string(1 to 3);
  constant NoInputGates : string3x2 := ( "GND", "VCC" );

  type string6x6 is array(0 to 4) of string(1 to 6);
  constant BufferGates : string6x6 := (
    "BUFD  ", "BUFF  ", "CLKINT", "INV   ", "INVD  ");

  type EnumGates2 is (
    AND2,  AND2A,  AND2B,
    NAND2, NAND2A, NAND2B,
    NOR2,  NOR2A,  NOR2B,
    OR2,   OR2A,   OR2B,
    XNOR2, XOR2 );

  type EnumGates3 is (
     AND3,  AND3A, AND3B,  AND3C,  AO1,    AO12,  AO13,   AO14,  AO15,  AO16,
     AO17,  AO18,  AO1A,   AO1B,   AO1C,   AO1D,  AO1E,   AOI1,  AOI1A, AOI1B,
     AOI1C, AOI1D, AOI5,   AX1,    AX1A,   AX1B,  AX1C,   AX1D,  AX1E,  AXO1,
     AXO2,  AXO3,  AXO5,   AXO6,   AXO7,   AXOI1, AXOI2,  AXOI3, AXOI4, AXOI5,
     AXOI7, MAJ3,  MAJ3X,  MAJ3XI, MIN3,   MIN3X, MIN3XI, MX2,   MX2A,  MX2B,
     MX2C,  NAND3, NAND3A, NAND3B, NAND3C, NOR3,  NOR3A,  NOR3B, NOR3C, OA1,
     OA1A,  OA1B,  OA1C,   OAI1,   OR3,    OR3A,  OR3B,   OR3C,  XA1,   XA1A,
     XA1B,  XA1C,  XAI1,   XAI1A , XO1,    XO1A,  XOR3,   XNOR3, ZOR3,  ZOR3I);

  -- ALL THE Dxxxx gates !
  type string8x88 is array(1 to 88) of string(1 to 8);
  constant EnumDFF : string8x88 := (
    "DFN0    ", "DFN1    ", "DFI0    ", "DFI1    ",
    "DFN0C0  ", "DFN1C0  ", "DFI0C0  ", "DFI1C0  ", 
    "DFN0C1  ", "DFN1C1  ", "DFI0C1  ", "DFI1C1  ",
    "DFN0E0  ", "DFN1E0  ", "DFI0E0  ", "DFI1E0  ",
    "DFN0E0C0", "DFN1E0C0", "DFI0E0C0", "DFI1E0C0",
    "DFN0E0C1", "DFN1E0C1", "DFI0E0C1", "DFI1E0C1",
    "DFN0E0P0", "DFN1E0P0", "DFI0E0P0", "DFI1E0P0",
    "DFN0E0P1", "DFN1E0P1", "DFI0E0P1", "DFI1E0P1",
    "DFN0E1  ", "DFN1E1  ", "DFI0E1  ", "DFI1E1  ",
    "DFN0E1C0", "DFN1E1C0", "DFI0E1C0", "DFI1E1C0",
    "DFN0E1C1", "DFN1E1C1", "DFI0E1C1", "DFI1E1C1",
    "DFN0E1P0", "DFN1E1P0", "DFI0E1P0", "DFI1E1P0",
    "DFN0E1P1", "DFN1E1P1", "DFI0E1P1", "DFI1E1P1",
    "DFN0P0  ", "DFN1P0  ", "DFI0P0  ", "DFI1P0  ",
    "DFN0P1  ", "DFN1P1  ", "DFI0P1  ", "DFI1P1  ",
    "DFN0P1C1", "DFN1P1C1", "DFI0P1C1", "DFI1P1C1",
    "DLI0    ", "DLI1    ", "DLN0    ", "DLN1    ",
    "DLI0C0  ", "DLI1C0  ", "DLN0C0  ", "DLN1C0  ",
    "DLI0C1  ", "DLI1C1  ", "DLN0C1  ", "DLN1C1  ",
    "DLI0P0  ", "DLI1P0  ", "DLN0P0  ", "DLN1P0  ",
    "DLI0P1  ", "DLI1P1  ", "DLN0P1  ", "DLN1P1  ",
    "DLI0P1C1", "DLI1P1C1", "DLN0P1C1", "DLN1P1C1");

  function SLV_to_bin(s:std_logic_vector) return string;

  impure function Gate_Census return integer;
  procedure update_select_gate(gate_number : integer; bit_number : integer; flip : integer);
  impure function InferLUT(L: std_logic_vector; s: string;
     gate_number : integer; lastbit : integer; exclude: std_logic_vector) return std_logic_vector;
  procedure update_histogram;
  procedure display_histogram;

  impure function LookItUp1 (A       : std_logic; LUT2 : LookupType2; gate_number : integer) return std_logic;
  impure function LookItUp2 (A, B    : std_logic; LUT4 : LookupType4; gate_number : integer) return std_logic;
  impure function LookItUp3 (A, B, C : std_logic; LUT8 : LookupType8; gate_number : integer) return std_logic;

  function ComputeLUT4 (t: EnumGates2) return LookupType4;
  function ComputeLUT8 (t: EnumGates3) return LookupType8;
end PA3_genlib;

package body PA3_genlib is

  -- from YGREC8/YGREC8_utils.vhdl:
  function SLV_to_bin(s:std_logic_vector) return string is
    variable t : string(s'range);
    variable u : string(3 downto 1);
  begin
    for i in s'range loop
      u := std_logic'image(s(i));
      t(i) := u(2);
    end loop;
    return t;
  end SLV_to_bin;

  -- just count the number of registered gates
  impure function Gate_Census return integer is
  begin
    gate_instance_counter := gate_instance_counter + 1;
    return gate_instance_counter;
  end Gate_Census;

  -- now that we know how many gates we have,
  -- allocate and initialise some shared variables with the generics,
  --   as well as the netlist/histograms
  procedure update_select_gate(
    gate_number : integer; bit_number : integer; flip : integer) is
--    variable i, j, k : integer;
  begin
    report integer'image(gate_instance_counter) & "x A3P gates found.";
    if gate_number < 1 then
      report " gate number: " & integer'image(gate_number)
           & "  bit number: " & integer'image( bit_number);
      report " -- no gate function is altered --";
      report "List of the registered gates:";
    end if;
    gate_select_number := gate_number;
     bit_select_number :=  bit_number;
              flip_LUT := flip;
              run_mode := trace;  -- To be changed later
    HistoPtr := new type_Histo_Array(1 to gate_instance_counter);
    indexPtr := new integer_array(0 to gate_instance_counter);
  end update_select_gate;

  -- This function is called at the initialisation time of the gate,
  -- after the generics are evaluated and Gate_Census gives the gate_number,
  -- but before the clock starts.
  -- If fast/default mode, the arrays are not initialised and LUT is simply copied.
  -- Otherwise, this allows us to list the gate or initialise and/or tweak its LUT.
  impure function InferLUT(L: std_logic_vector; s: string;
       gate_number : integer; lastbit : integer; exclude: std_logic_vector)
  return std_logic_vector is
    variable T : std_logic_vector(L'range) := L;
    variable i : integer;
  begin
    if run_mode /= fast then
      if gate_select_number < 1 then  -- just a listing
        report " gate#" & integer'image(gate_number) & " : " & s;
      end if;

      -- clear the histogram :
      for i in type_Histo8'range loop
        HistoPtr(gate_number).histo(i) := 0;
      end loop;

      -- register an exclusion :
      if exclude'length > 0 then
        for i in exclude'range loop
          exit when i > lastbit;
          if exclude(i) = 'X' then
            T(i             ) := 'X';
            T(i + lastbit+ 1) := 'X';
            HistoPtr(gate_number).histo(i             ):= -1;
            HistoPtr(gate_number).histo(i + lastbit+ 1):= -1;
          end if;
        end loop;
        report " Excluding inputs of gate #" & integer'image(gate_number) & " -> " & SLV_to_bin(T);
      end if;

      -- if flip or trace :
      if gate_select_number >= 0 then
        -- init the histogram:
        HistoPtr(gate_number).last := lastbit;

        -- register into the list of updates/display:
        if lastbit > 0 then
          indexPtr(registered_instance_counter) := gate_number;
          registered_instance_counter := registered_instance_counter + 1;
        end if;

        if  gate_select_number = gate_number
         and bit_select_number >= 0
         and bit_select_number <= lastbit then

          assert T(bit_select_number) /= 'X'
             report "Can't flip unreachable bit of gate#" & integer'image(gate_number)
             severity failure;

          if flip_LUT = 1 then
            -- Alter the selected bit
            i := bit_select_number + lastbit+ 1;
            if T(bit_select_number) = '0' then
              T(i) := 'H'; -- invert the high/shadow bits
              T(bit_select_number) := '1';
            else
              T(i) := 'L';
              T(bit_select_number) := '0';
            end if;
            report " Flipping bit " & integer'image(bit_select_number)
                 & " of gate #" & integer'image(gate_number) & " -> " & SLV_to_bin(T);
          else
            -- transform the selected value into a meta value that will propagate
            if T(bit_select_number) = '0' then
              T(bit_select_number) := 'L';
            else
              T(bit_select_number) := 'H';
            end if;
            report " Tracing bit " & integer'image(bit_select_number)
               & " of gate #" & integer'image(gate_number) & " -> " & SLV_to_bin(T);
          end if;
        end if;
      end if;
    end if;
    return T;
  end InferLUT;

  procedure update_histogram is
    variable i, j, k : integer;
    variable l : big_sint_t;
    variable s, t : std_logic;
  begin
    assert cycle_counter < integer'high
       report "Simulation reached maximum cycle count"
       severity failure;
    cycle_counter := cycle_counter + 1;
    i := 0;
    while i < registered_instance_counter loop
      j := indexPtr(i);
      k := HistoPtr(j).index;
      l := HistoPtr(j).histo(k);
      assert l /= -1
        report "Gate#" & integer'image(j) & " reached excluded state '" & integer'image(k) & "'"  
        severity failure ;
      HistoPtr(j).histo(k) := l + 1;

      -- count the toggles at the output:
         s := HistoPtr(j).curOut;
      if s /= HistoPtr(j).prevOut then
        HistoPtr(j).changes := HistoPtr(j).changes + 1;
      end if;
      HistoPtr(j).prevOut := s;

      i := i+1;
    end loop;
  end update_histogram;

  procedure display_histogram is
    variable i, j, k, m, n : integer;
    variable c : big_uint_t;
    variable o : big_sint_t;
    variable l : line;
  begin
    report "End of simulation ! " & integer'image(cycle_counter) & " cycles.";

    c := 0;

    i := 0;
    while i < registered_instance_counter loop
      j := indexPtr(i);
      m := HistoPtr(j).last;
      n := m+m+1;
      for k in 0 to n loop
        o := HistoPtr(j).histo(k);
        if o = -1 then
          write(l, "^ " & empty_str);
        else
          write(l, big_sint_t'image(o) & " ");
        end if;
        if k = m then
          write(l, " |  " & empty_str);
        end if;
      end loop;

      c := c + HistoPtr(j).changes;

      report " h(" & integer'image(j) & ") = " & L.all
            & " (" & big_sint_t'image(HistoPtr(j).histo(special))
            & ") [" & big_uint_t'image(HistoPtr(j).changes) & "]";
      Deallocate(L);
      i := i+1;
    end loop;

    report "Total state changes: " & big_uint_t'image(c) & " (not counting fanout)";
  end display_histogram;

  impure function LookItUp1 (A : std_logic; LUT2 : LookupType2;
          gate_number : integer) return std_logic is
    variable index : integer := 0;
    variable R : std_logic := A;
  begin

    if run_mode = fast then
      --  from LookItUp1_fast:
      if A = '1' then
        return LUT2(1);
      else
        return LUT2(0);
      end if;
    end if;

    if run_mode = trace then
      case A is
        when '0'    => index := 0;
        when '1'    => index := 1;
        when 'L'    => index := 2;
        when 'H'    => index := 3;
        when others => index := SPECIAL;
              HistoPtr(gate_number).index := index;
              return R;
      end case;
      R := LUT2(index);
      HistoPtr(gate_number).index := index;
      HistoPtr(gate_number).curOut := R;
      return R;
    end if;

    assert false report "invalid mode !" severity failure;
  end LookItUp1;

  impure function LookItUp2 (A, B : std_logic; LUT4 : LookupType4;
          gate_number : integer) return std_logic is
    variable index, meta : integer := 0;
    variable R : std_logic := A;
  begin

    if run_mode = fast then
      if A = '1' then
        index := 2;
      end if;
      if B = '1' then
        index := index+1;
      end if;
      return LUT4(index);
    end if;

    if run_mode = trace then
      case A is
        when '0'    =>  NULL;
        when '1'    =>  index := 2;
        when 'L'    =>              meta := 4;
        when 'H'    =>  index := 2; meta := 4;
        when others =>  HistoPtr(gate_number).index := SPECIAL; return A;
      end case;

      case B is
        when '0'    =>  NULL;
        when '1'    =>  index := index+1;
        when 'L'    =>                    meta := 4;       
        when 'H'    =>  index := index+1; meta := 4;
        when others =>  HistoPtr(gate_number).index := SPECIAL; return B;
      end case;

      index := index+meta;
      R := LUT4(index);
      HistoPtr(gate_number).index := index;
      HistoPtr(gate_number).curOut := R;
      return R;
    end if;

    assert false report "invalid mode !" severity failure;
  end LookItUp2;

  impure function LookItUp3 (A, B, C : std_logic; LUT8 : LookupType8;
      gate_number : integer) return std_logic is
    variable index, meta : integer := 0;
    variable R : std_logic := A;
  begin
    if run_mode = fast then
      if A = '1' then
        index := 4;
      end if;

      if B = '1' then
        index := index+2;
      end if;

      if C = '1' then
        index := index+1;
      end if;
      return LUT8(index);
    end if;

    if run_mode = trace then
      case A is
        when '0'    =>  NULL;
        when '1'    =>  index := 4;
        when 'L'    =>              meta := 8;
        when 'H'    =>  index := 4; meta := 8;
        when others =>  HistoPtr(gate_number).index := SPECIAL; return A;
      end case;

      case B is
        when '0'    =>  NULL;
        when '1'    =>  index := index+2;
        when 'L'    =>                    meta := 8;
        when 'H'    =>  index := index+2; meta := 8;
        when others =>  HistoPtr(gate_number).index := SPECIAL; return B;
      end case;

      case C is
        when '0'    =>  NULL;
        when '1'    =>  index := index+1;
        when 'L'    =>                    meta := 8;
        when 'H'    =>  index := index+1; meta := 8;
        when others =>  HistoPtr(gate_number).index := SPECIAL; return C;
      end case;

      index := index+meta;
      R := LUT8(index);
      HistoPtr(gate_number).index := index;
      HistoPtr(gate_number).curOut := R;
      return R;
    end if;

    assert false report "invalid mode !" severity failure;
  end LookItUp3;

-----------------------------------------------------------------------------------------
-- only used by gategen.vhdl :

  function ComputeLUT4 (t: EnumGates2) return LookupType4 is
    variable L : LookupType4;
    variable A, B, C, Y : std_logic;
    variable i : integer;
  begin
    i := 0;

    A := '0';
    loop

      B := '0';
      loop
        case t is
          when AND2   => Y :=          A  and      B  ;
          when AND2A  => Y :=     (not A) and      B  ;
          when AND2B  => Y :=     (not A) and (not B) ;
          when NAND2  => Y := not(     A  and      B) ;
          when NAND2A => Y := not((not A) and      B) ;
          when NAND2B => Y := not((not A) and (not B));
          when NOR2   => Y :=          A  nor      B  ;
          when NOR2A  => Y := not((not A)  or      B) ;
          when NOR2B  => Y := not((not A)  or (not B));
          when OR2    => Y :=          A   or      B  ;
          when OR2A   => Y :=     (not A)  or      B  ;
          when OR2B   => Y :=     (not A)  or (not B) ;
          when XNOR2  => Y := not     (A  xor      B) ;
          when XOR2   => Y :=          A  xor      B  ;
          when others => report "Function not implemented : "
                  & EnumGates2'image(t) severity failure;
        end case;

        L(i) := Y;
        if Y = '1' then
          L(i+4) := 'H';
        else
          L(i+4) := 'L';
        end if;
        i := i+1;

        exit when B = '1'; 
        B := '1';
      end loop;

      exit when A = '1';
      A := '1';
    end loop;

    return L;
  end ComputeLUT4;


  function ComputeLUT8 (t: EnumGates3) return LookupType8 is
    variable L : LookupType8;
    variable A, B, C, Y : std_logic;
    variable i : integer;
  begin
    i := 0;

    A := '0';
    loop

      B := '0';
      loop 

        C := '0';
        loop

          case t is
            when AND3   => Y := (     A  and      B ) and      C ;
            when AND3A  => Y := ((not A) and      B ) and      C ;
            when AND3B  => Y := ((not A) and (not B)) and      C ;
            when AND3C  => Y := ((not A) and (not B)) and (not C);
            when AO12   => Y := ((not A) and B) or ((not A) and (not C)) or (B and (not C)) or (A and (not B) and C);
            when AO13   => Y := (A and B) or (A and (not C)) or (B and (not C));
            when AO14   => Y := (A and B) or (A and (not C)) or (B and (not C)) or ((not A) and (not B) and C);
            when AO15   => Y := (A and (not B) and C) or ((not A) and B and C) or ((not A) and (not B) and (not C));
            when AO16   => Y := (A and B and (not C)) or ((not A) and (not B) and C);
            when AO17   => Y := (A and B and C) or ((not A) and B and (not C)) or ((not A) and (not B) and C);
            when AO18   => Y := ((not A) and B) or ((not A) and (not C)) or (B and (not C));
            when AO1    => Y :=      (     A  and      B ) or      C;
            when AO1A   => Y :=      ((not A) and      B ) or      C;
            when AO1B   => Y :=      (     A  and      B ) or (not C);
            when AO1C   => Y :=      ((not A) and      B ) or (not C);
            when AO1D   => Y :=      ((not A) and (not B)) or      C;
            when AO1E   => Y :=      ((not A) and (not B)) or (not C);
            when AOI1   => Y := not ((     A  and      B ) or      C );
            when AOI1A  => Y := not (((not A) and      B ) or      C );
            when AOI1B  => Y := not ((     A  and      B ) or (not C));
            when AOI1C  => Y := not (((not A) and (not B)) or      C );
            when AOI1D  => Y := not (((not A) and (not B)) or (not C));
            when AOI5   => Y := not (((not A) and B and C) or (A and (not B) and (not C)) );
            when AX1    => Y :=     ((not A) and      B ) xor C;
            when AX1A   => Y := not(((not A) and      B ) xor C);
            when AX1B   => Y :=     ((not A) and (not B)) xor C;
            when AX1C   => Y :=     (     A  and      B ) xor C;
            when AX1D   => Y := not(((not A) and (not B)) xor C);
            when AX1E   => Y := not((     A  and      B ) xor C);
            when AXO1   => Y :=     (     A  and      B ) or (     B  xor C);
            when AXO2   => Y :=     ((not A) and      B ) or (     B  xor C);
            when AXO3   => Y :=     (     A  and (not B)) or (     B  xor C);
            when AXO5   => Y :=     ((not A) and      B ) or ((not B) xor C);
            when AXO6   => Y :=     (     A  and (not B)) or ((not B) xor C);
            when AXO7   => Y :=     ((not A) and (not B)) or (     B  xor C);
            when AXOI1  => Y := not((     A  and      B ) or (     B  xor C));
            when AXOI2  => Y := not(((not A) and      B ) or (     B  xor C));
            when AXOI3  => Y := not((     A  and (not B)) or (     B  xor C));
            when AXOI4  => Y := not((     A  and      B ) or ((not B) xor C));
            when AXOI5  => Y := not(((not A) and      B ) or ((not B) xor C));
            when AXOI7  => Y := not(((not A) and (not B)) or (     B  xor C));
            when MAJ3   => Y :=     (A and B            ) or (A and             C) or (            B and C);
            when MAJ3X  => Y :=     (A and B and (not C)) or (A and (not B) and C) or ((not A) and B and C);
            when MAJ3XI => Y := not((A and B and (not C)) or (A and (not B) and C) or ((not A) and B and C));
            when MIN3   => Y :=     ((not A) and (not B)      ) or ((not A) and       (not C)) or (      (not B) and (not C));
            when MIN3X  => Y :=     ((not A) and (not B) and C) or ((not A) and B and (not C)) or (A and (not B) and (not C));  --- not A3P ! 54SX, 54SX-A, 54SX-S, eX
            when MIN3XI => Y := not(((not A) and (not B) and C) or ((not A) and B and (not C)) or (A and (not B) and (not C)));
            when MX2    => if C='0' then  Y :=      A ; else Y :=      B ; end if;
            when MX2A   => if C='0' then  Y := (not A); else Y :=      B ; end if;
            when MX2B   => if C='0' then  Y :=      A ; else Y := (not B); end if;
            when MX2C   => if C='0' then  Y := (not A); else Y := (not B); end if;
            when NAND3  => Y := not ((     A  and      B ) and      C );
            when NAND3A => Y := not (((not A) and      B ) and      C );
            when NAND3B => Y := not (((not A) and (not B)) and      C );
            when NAND3C => Y := not (((not A) and (not B)) and (not C));
            when NOR3   => Y := not ((    A  or      B) or      C);
            when NOR3A  => Y := not ((not A) or      B  or      C );
            when NOR3B  => Y := not ((not A) or (not B) or      C );
            when NOR3C  => Y := not ((not A) or (not B) or (not C));
            when OA1    => Y := (     A  or B) and      C ;
            when OA1A   => Y := ((not A) or B) and      C ;
            when OA1B   => Y := (     A  or B) and (not C);
            when OA1C   => Y := ((not A) or B) and (not C);
            when OAI1   => Y := not ((A or B) and C);
            when OR3    => Y := (A or B) or C;
            when OR3A   => Y := (not A) or      B  or      C ;
            when OR3B   => Y := (not A) or (not B) or      C ;
            when OR3C   => Y := (not A) or (not B) or (not C);
            when XA1    => Y :=          (A xor B)  and      C;
            when XA1A   => Y :=     (not (A xor B)) and      C;
            when XA1B   => Y :=          (A xor B)  and (not C);
            when XA1C   => Y :=     (not (A xor B)) and (not C);
            when XAI1   => Y := not(     (A xor B)  and	     C);
            when XAI1A  => Y := not((not (A xor B)) and      C);
            when XNOR3  => Y :=  not ((A xor B) xor C);
            when XO1    => Y :=      (A xor B)  or C;
            when XO1A   => Y := (not (A xor B)) or C;
            when XOR3   => Y :=      (A xor B) xor C;
            when ZOR3   => Y :=      (A and B and C) or ((not A) and (not B) and (not C));
            when ZOR3I  => Y := not ((A and B and C) or ((not A) and (not B) and (not C)));
            when others => report "Function not implemented : "
                                & EnumGates3'image(t) severity failure;
          end case;

          L(i) := Y;
          if Y = '1' then
            L(i+8) := 'H';
          else
            L(i+8) := 'L';
          end if;
          i := i+1;

          exit when C = '1';
          C := '1';
        end loop;

        exit when B = '1'; 
        B := '1';
      end loop;

      exit when A = '1';
      A := '1';
    end loop;

    return L;
  end ComputeLUT8;

end PA3_genlib;
