-- gategen.vhdl
-- created jeu. août  8 10:18:01 CEST 2019  by Yann Guidon (whygee@f-cpu.org)
-- version sam. août 10 08:06:58 CEST 2019
-- version dim. août 11 17:36:51 CEST 2019
-- version lun. août 12 18:36:04 CEST 2019
-- version mar. août 13 18:05:15 CEST 2019
-- version dim. août 25 10:25:30 CEST 2019 : DLxxxx added
-- version jeu. nov. 21 03:38:34 CET  2019 : removing "trace"vs"fast"
-- version ven. nov. 22 23:39:03 CET  2019 : generic gates
--
-- Released under the GNU AGPLv3 license
--
-- This file generates PA3_components.vhdl and PA3_definitions.vhdl.
--
-- TODO : reduce redundancies...

Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
Library work;
    use work.all;
    use work.PA3_genlib.all;
use STD.textio.all;

entity GateGen is
  generic ( gate_delay : string := "");
end GateGen;

architecture runme of GateGen is
  file gate_defs : text open write_mode is "PA3_definitions.vhdl";
  file gate_list : text open write_mode is "PA3_components.vhdl";
begin

  bench: process
    variable c, Epol, Cpol, Ppol, op : character;
    variable i, j, indent : integer := 0;
    variable spaces : string(1 to 12) := (others=>' '); 
    variable s : EnumGates2;
    variable t : EnumGates3;
    variable L2 : LookupType2;
    variable L4 : LookupType4;
    variable L8 : LookupType8;
    variable row : line;
    variable QN, StrE : string(1 to 2);
    variable GCLK : string(1 to 3);
    variable StrC, StrP : string(1 to 5);

    procedure wl(s: string) is
    begin
      write(row, s);
      writeline(gate_list, row);
    end wl;

    procedure wd(s: string) is
    begin
      write(row, s);
      writeline(gate_defs, row);
    end wd;

    procedure wdi(s: string) is
    begin
      write(row, spaces(1 to indent) & s);
      writeline(gate_defs, row);
    end wdi;

    procedure wdg is
    begin
      wd("  generic( Gate_Number : integer := Gate_Census;");
      wd("           exclude: std_logic_vector := """");");
    end wdg;

    procedure uselib is
    begin
      wd("");
      wd("--------------------------------------------------");
      wd("");
      wd("Library ieee;");
      wd("    use ieee.std_logic_1164.all;");
      wd("Library work;");
      wd("    use work.all;");
      wd("    use work.PA3_genlib.all;");
      wd("");
    end uselib;

    procedure shift_indent(i : integer) is
    begin
      if i >= 4 then
        indent := i;
      end if;
    end shift_indent;

  begin
    wl("-- File : PA3_components.vhdl");
    wl("-- DO NOT MODIFY ! Generated file !");
    wl("-- (C) Yann Guidon 20190808 - 20191123");
    wl("");
    wl("Library ieee;");
    wl("    use ieee.std_logic_1164.all;");
    wl("");
    wl("package proasic3 is");

    wd("-- File : PA3_definitions.vhdl");
    wd("-- DO NOT MODIFY ! Generated file !");
    wd("-- (C) Yann Guidon 20190808 - 20191123");
    if gate_delay = "" then
      report "No gate delay specified. Change this value with -ggate_delay=""1 ns"" ";
    else
      report "Gate delay set to " & gate_delay;
    end if;

    wl("");
    wl("  component register_generics port(gate_select_number, ");
    wl("     bit_select_number, flip_LUT : in integer; clock : in std_logic); end component;"); 
    wl("");

    uselib;
    wd("-- Note if you use the modes trace, alter, netlist :");
    wd("-- /!\ You MUST write this in front/before including or instanciating the other gates !");
    wd("-- It must be called first to initialise important values from the generics /!\");
    wd("entity register_generics is");
    wd("  port( gate_select_number, bit_select_number, flip_LUT: in integer;");
    wd("        clock : in std_logic);");
    wd("end register_generics;");
    wd("");
    wd("architecture rtl of register_generics is");
    wd("begin");
    wd("  update_select_gate(gate_select_number, bit_select_number, flip_LUT); -- called only once at start");
    wd("");
    wd("  process(clock) begin");
    wd("    case clock is");
    wd("      when '1' => update_histogram;");
    wd("      when 'X' => display_histogram;");
    wd("      when others => NULL;");
    wd("    end case;");
    wd("  end process;");
    wd("end rtl;");

    wd("");
    wd("--------------------------------------------------");
    wd("");
    wd("-- Generic gates :");

    j := 2;
    for i in 1 to 3 loop
      write(row,"  component generic_gate" & integer'image(i) & " generic(exclude: std_logic_vector); port(A");
      if i > 1 then  write(row,", B" & empty_str); else  write(row,"   " & empty_str); end if;
      if i > 2 then  write(row,", C" & empty_str); else  write(row,"   " & empty_str); end if;
      wl(" : in std_logic; LUT : in std_logic_vector; Y : out std_logic); end component;");

      uselib;
      wd("entity generic_gate" & integer'image(i) & " is");
      wdg;
      write(row,"  port( A" & empty_str);
      if i > 1 then  write(row,", B" & empty_str); end if;
      if i > 2 then  write(row,", C" & empty_str); end if;
      wd(" : in std_logic;");
      wd("        LUT : in std_logic_vector;");
      wd("        Y : out std_logic);");
      wd("end generic_gate" & integer'image(i) & ";");
      wd("");
      wd("architecture trace of generic_gate" & integer'image(i) & " is");
      wd("  signal LUT" & integer'image(j) & " : LookupType" & integer'image(j) & ";");
      wd("begin");
      wd("  LUT" & integer'image(j) & " <= InferLUT(LUT, LUT" & integer'image(j)
           &  "'instance_name, Gate_Number, "  & integer'image(j-1) & ", exclude);");    -- called only once at start
      write(row, "  Y <= LookItUp" & integer'image(i) & "(A" & empty_str);
      if i > 1 then  write(row,", B" & empty_str); end if;
      if i > 2 then  write(row,", C" & empty_str); end if;
      write(row, ", LUT" & integer'image(j) & ", Gate_Number)" & empty_str);  -- kludge

      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");
      wd("end trace;");

      j := j*2;
    end loop;

    wd("");
    wd("--------------------------------------------------");
    wd("");
    wd("-- 0-input gates :");

    wl("");
    wl("-- 0-input gates :");

    for i in NoInputGates'range loop
      wl("  component "& NoInputGates(i) & " generic(exclude: std_logic_vector); port(Y : out std_logic); end component;");

      uselib;
      wd("entity " & NoInputGates(i) & " is");
      wdg;
      wd("  port(Y : out std_logic);");
      wd("end " & NoInputGates(i) & ";");

      wd("");
      wd("architecture trace of " & NoInputGates(i) & " is");
      wd("  signal LUT1 : LookupType1;");
      wd("begin");
      wd("  LUT1 <= InferLUT(""" & integer'image(i) & integer'image(i)
               &  """, LUT1'instance_name, Gate_Number, 0, exclude);");
      wd("  Y <= LUT1(0);");
      wd("end trace;");
    end loop;

    wl("");
    wl("-- 1-input gates :");
    
    for i in BufferGates'range loop
      wl("  component "& BufferGates(i) & " generic(exclude: std_logic_vector); port(A : in std_logic; Y : out std_logic); end component;");

      if BufferGates(i)(1) = 'I' then
        L2:="10HL"; -- inverter gate
      else
        L2:="01LH"; -- note: ascending order of the bits !
      end if;
      uselib;
      wd("entity " & BufferGates(i) & " is");
      wdg;
      wd("  port( A : in std_logic;");
      wd("        Y : out std_logic);");
      wd("end " & BufferGates(i) & ";");

      wd("");
      wd("architecture trace of " & BufferGates(i) & " is");
      wd("  signal LUT2 : LookupType2;");
      wd("begin");
      wd("  LUT2 <= InferLUT(""" & SLV_to_bin(L2)
         & """, LUT2'instance_name, Gate_Number, 1, exclude);");
      write(row, "  Y <= LookItUp1(A, LUT2, Gate_Number)" & empty_str);  -- kludge

      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");

      wd("end trace;");
    end loop;


    wl("");
    wl("-- 2-inputs gates :");

    for s in EnumGates2 loop
      L4 := ComputeLUT4(s);
      write(row, "  component " & EnumGates2'image(s), left, 19);
      wl(" generic(exclude: std_logic_vector); port(A, B : in std_logic; Y : out std_logic); end component; -- " & SLV_to_bin(L4(0 to 3)));

      uselib;
      wd("entity " & EnumGates2'image(s) & " is");
      wdg;
      wd("  port( A, B : in std_logic;");
      wd("           Y : out std_logic);");
      wd("end " & EnumGates2'image(s) & ";");

      wd("");
      wd("architecture trace of " & EnumGates2'image(s) & " is");
      wd("  signal LUT4 : LookupType4;");
      wd("begin");
      wd("  LUT4 <= InferLUT(""" & SLV_to_bin(L4)
         & """, LUT4'instance_name, Gate_Number, 3, exclude);");
      write(row, "  Y <= LookItUp2(A, B, LUT4, Gate_Number)" & empty_str);  -- kludge

      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");

      wd("end trace;");
    end loop;

    wl("");
    wl("-- 3-inputs gates :");
    for t in EnumGates3 loop
      case t is
        when MX2  |
             MX2A |
             MX2B |
             MX2C   =>  op := 'S';
        when others =>  op := 'C';
      end case;

      L8 := ComputeLUT8(t);
      write(row, "  component " & EnumGates3'image(t), left, 19);
      wl(" generic(exclude: std_logic_vector); port(A, B, " & op
          & " : in std_logic; Y : out std_logic); end component; -- " & SLV_to_bin(L8(0 to 7)));

      uselib;
      wd("entity " & EnumGates3'image(t) & " is");
      wdg;
      wd("  port( A, B, " & op &" : in std_logic;");
      wd("              Y : out std_logic);");
      wd("end " & EnumGates3'image(t) & ";");

      wd("");
      wd("architecture trace of " & EnumGates3'image(t) & " is");
      wd("  signal LUT8 : LookupType8;");
      wd("begin");
      wd("  LUT8 <= InferLUT(""" & SLV_to_bin(L8)
          & """, LUT8'instance_name, Gate_Number, 7, exclude);");    -- called only once at start
      write(row, "  Y <= LookItUp3(A, B, " & op & ", LUT8, Gate_Number)" & empty_str);  -- kludge

      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");
      wd("end trace;");
    end loop;

    wl("");
    wl("-- Sequential gates :");

    for i in EnumDFF'range loop

      if EnumDFF(i)(2)='F' then
        GCLK := "CLK";
      else
        GCLK := "G  ";
      end if;

      if EnumDFF(i)(3) = 'I' then
        QN := "QN";
        L2:="01LH"; -- inverter
      else
        QN := " Q";
        L2:="10HL";
      end if;

      StrE := "  ";
      j := 5;
      if EnumDFF(i)(5) = 'E' then
        StrE := "E,";
        j := 7;
        Epol := EnumDFF(i)(6);
      end if;

      StrP := "     ";
      if EnumDFF(i)(j) = 'P' then
        StrP := ", PRE";
        Ppol := EnumDFF(i)(j+1);
        j := j+2;
      end if;

      StrC := "     ";
      if j < 8 and EnumDFF(i)(j) = 'C' then
        StrC := ", CLR";
        Cpol := EnumDFF(i)(j+1);
      end if;

      wl("  component " & EnumDFF(i) & " generic(exclude: std_logic_vector); port(D, " & StrE & " " & GCLK & StrP & StrC
         & " : in std_logic; " & QN &" : out std_logic); end component;");

      uselib;
      wd("entity " & EnumDFF(i) & " is");
      wdg;
      wd("  port(D, " & StrE & " " & GCLK & StrP & StrC & " : in std_logic;");
      wd("       " & QN & " : out std_logic);");
      wd("end " & EnumDFF(i) & ";");
      wd("");
      wd("architecture trace of " & EnumDFF(i) & " is");
      wd("  signal LUT2 : LookupType2;");
      wd("begin");
      wd("  LUT2 <= InferLUT(""" & SLV_to_bin(L2)
         & """, LUT2'instance_name, Gate_Number, 1, exclude);");

      write(row, "  process(" & GCLK & empty_str);
      if StrE(1) = 'E' then
        write(row, ", E" & empty_str);
      end if;
      if StrP(1) = ',' then
        write(row, StrP);
      end if;
      if StrC(1) = ',' then
        write(row, StrC);
      end if;
      wd(")");
      wd("  begin");

      shift_indent(4);

      if StrC(1) = ',' then
        wdi("if CLR='" & Cpol & "' then");
        write(row, spaces(1 to indent) & "  "& QN &" <= '0'");
        if gate_delay /= "" then
          write(row, " after " & gate_delay);
        end if;
        wd(";");
        wdi("else");
        shift_indent(6);
      end if;

        -- Clear has precedence over Preset
        if StrP(1) = ',' then
          wdi("if PRE='" & Ppol & "' then");
          write(row, spaces(1 to indent) & "  "& QN &" <= '1'");
          if gate_delay /= "" then
            write(row, " after " & gate_delay);
          end if;
          wd(";");
          wdi("else");
          shift_indent(indent+2);
        end if;

          write(row, spaces(1 to indent) & "if ");

          if EnumDFF(i)(2)='F' then -- Flip Flop
            if EnumDFF(i)(4)='1' then
              write(row, "rising" & empty_str);
            else
              write(row, "falling" & empty_str);
            end if;
            write(row, "_edge(CLK)" & empty_str);
            if StrE(1) = 'E' then
              write(row, " and E='" & Epol & "'");
            end if;
          else         -- Latch
            write(row, "G='" & EnumDFF(i)(4) & "'");
          end if;

          wd(" then");
          write(row, spaces(1 to indent) & "  " & QN & " <= " &
                      "LookItUp1(D , LUT2, Gate_Number)" & empty_str);

          if gate_delay /= "" then
            write(row, " after " & gate_delay);
          end if;
          wd(";");
          wdi("end if;");

        if StrP(1) = ',' then
          shift_indent(indent-2);
          wdi("end if;");
        end if;

      if StrC(1) = ',' then
        shift_indent(indent-2);
        wdi("end if;");
      end if;

      wd("  end process;");
      wd("end trace;");
    end loop;

    wl("");
    wl("  component RAM4K9 port(");
    wl("    ADDRA11, ADDRA10, ADDRA9, ADDRA8, ADDRA7, ADDRA6,");
    wl("    ADDRA5,  ADDRA4,  ADDRA3, ADDRA2, ADDRA1, ADDRA0,");
    wl("    ADDRB11, ADDRB10, ADDRB9, ADDRB8, ADDRB7, ADDRB6,");
    wl("    ADDRB5,  ADDRB4,  ADDRB3, ADDRB2, ADDRB1, ADDRB0,");
    wl("    DINA8, DINA7, DINA6, DINA5, DINA4, DINA3, DINA2, DINA1, DINA0,");
    wl("    DINB8, DINB7, DINB6, DINB5, DINB4, DINB3, DINB2, DINB1, DINB0,");
    wl("    WIDTHA0, WIDTHA1, WIDTHB0, WIDTHB1, PIPEA, PIPEB, WMODEA, WMODEB,");
    wl("    BLKA, BLKB, WENA, WENB, CLKA, CLKB, RESET : in std_logic := 'U';");
    wl("    DOUTA8, DOUTA7, DOUTA6, DOUTA5, DOUTA4, DOUTA3, DOUTA2, DOUTA1, DOUTA0,");
    wl("    DOUTB8, DOUTB7, DOUTB6, DOUTB5, DOUTB4, DOUTB3, DOUTB2, DOUTB1, DOUTB0 : out std_logic);");
    wl("  end component;");
    wl("");
    wl("end proasic3;");

    wait;
  end process;

end runme;
