-- YGREC8/ALU8_tb.vhdl
-- version lun. nov. 13 01:43:33 CET 2017 by Yann Guidon (whygee@f-cpu.org)
-- version jeu. nov. 16 02:21:35 CET 2017 : uses SLV shorthands from YGREC8/YGREC8_def.vhdl
-- version jeu. mars 21 08:51:50 CET 2019 : renamed SRC & DST, test SET/pass
-- version lun. mars 25 06:10:20 CET 2019 : all opcodes pass
-- version dim. nov. 24 18:21:18 CET 2019 : adding fault injection
-- Released under the GNU AGPLv3 license

Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
Library work;
    use work.all;
    use work.ygrec8_def.all;
    use work.ygrec8_utils.all;
-- added :
Library proasic3;
    use proasic3.all;

entity ALU8_tb is
-- added :
  generic (
    -- for the A3P library :
    gate_select_number: integer := 0;  -- when >= 1, select a gate to alter
    bit_select_number: integer := -1;  -- select which input bit code to alter
    flip: integer := -1
       -- default : do nothing, don't instanciate register_generics
       -- clear to 0 to alter 0/1 to meta L/H code
       -- set to 1 to bit flip one bit of the LUT
  );
end ALU8_tb;

architecture plip of ALU8_tb is
  signal --CarryFlag,
      Cin, Cout, NEG, PassMask, OrXor, Rop2Mx, CMPS : SL := '0';
  signal F : SLV4 := "0000";
  signal SRI, SND, ROP2, SUM : SLV8 := "00000000";
  signal Inst : SLV16 := "0000000000000000";
-- added :
  signal clock : std_logic := '0';
begin
-- added :
  extra: if flip >= 0 generate
    -- this entity MUST be written first,
    -- at the very beginning of the toplevel file:
    rg: entity register_generics port map(
         gate_select_number => gate_select_number,
         bit_select_number => bit_select_number,
         flip_LUT=> flip, clock => clock);
  end generate;

  Inst <= F & "000000000000";

  tbdec: entity DEC_ALU8 port map (
    Inst      => Inst,
--    CarryFlag => CarryFlag,
    CMPS      => CMPS,
    Rop2Mx    => Rop2Mx,
    OrXor     => OrXor,
    PassMask  => PassMask,
    NEG       => NEG,
    Cin       => Cin);

  tb: entity ALU8 port map (
     CMPS     => CMPS,
     Rop2Mx   => Rop2Mx,
     OrXor    => OrXor,
     PassMask => PassMask,
     NEG      => NEG,
     Cin      => Cin,
     SRI      => SRI,
     SND      => SND,
     SUM      => SUM,
     ROP2     => ROP2,
     Cout     => Cout);

  bench: process
    variable j, k, j2, k2, errors: integer := 0;
    variable expected : SLV8;
    variable t : SLV10;
    variable C : SL;

    procedure ShowSignals(i: integer) is
    begin
      report -- " F=" & SLV_to_bin(std_logic_vector(F)) & " " & 
             OpcodeS_str_array(i)
           & " SRI=" & SLV_to_bin(SRI)
           & " SND=" & SLV_to_bin(SND)
           & " Cin=" &   std_logic'image(Cin)
           & " Cout=" &  std_logic'image(Cout)
           & " SUM=" & SLV_to_bin(SUM)
           & " ROP2=" & SLV_to_bin(ROP2)
           & " expect=" & SLV_to_bin(expected)
           & " Diff=" & SLV_to_bin(ROP2 XOR expected);
--           & " Diff=" & SLV_to_bin(t(8 downto 0) XOR (Cout & SUM));

      -- "early quit" to stop the test as soon as one
      -- error is found and keep run times short
      assert flip /= 1
         severity failure;
      errors:=errors+1;
    end ShowSignals;

  begin
-- added :
    if gate_select_number=-1 then
      wait;
    end if;

    for j in 0 to 255 loop
      SRI <= std_logic_vector(to_unsigned(j, 8));
      for k in 0 to 255 loop
        SND <= std_logic_vector(to_unsigned(k, 8));

        -- Bool:

        F <= Op_AND;
        clock <= '0'; wait for 10 ns;
        clock <= '1'; wait for 10 ns;
        expected := SRI AND SND;
        if expected /= ROP2 then ShowSignals(0); end if;

        F <= Op_OR;
        clock <= '0'; wait for 10 ns;
        clock <= '1'; wait for 10 ns;
        expected := SRI OR SND;
        if expected /= ROP2 then ShowSignals(1); end if;

        F <= Op_XOR;
        clock <= '0'; wait for 10 ns;
        clock <= '1'; wait for 10 ns;
        expected := SRI XOR SND;
        if expected /= ROP2 then ShowSignals(2); end if;

        F <= Op_ANDN;
        clock <= '0'; wait for 10 ns;
        clock <= '1'; wait for 10 ns;
        expected := SRI AND (NOT SND);
        if expected /= ROP2 then ShowSignals(3); end if;

        -- CMPU:
        F <= Op_CMPU;
        clock <= '0'; wait for 10 ns;
        clock <= '1'; wait for 10 ns;
        if k > j then C := '0'; else C := '1'; end if;
        if C /= Cout then ShowSignals(4); end if;

        -- CMPS:
        F <= Op_CMPS;
        clock <= '0'; wait for 10 ns;
        clock <= '1'; wait for 10 ns;
        if k > 127 then k2 := k-256; else k2 := k; end if;
        if j > 127 then j2 := j-256; else j2 := j; end if;
        if k2 > j2 then C := '0'; else C := '1'; end if;
        if C /= Cout then ShowSignals(5); end if;

        -- SUB:
        F <= Op_SUB;
        clock <= '0'; wait for 10 ns;
        clock <= '1'; wait for 10 ns;
        t := std_logic_vector(to_unsigned((j+512)-k, 10));
        expected := t(7 downto 0);
        if t(8 downto 0) /= ( not Cout & SUM) then ShowSignals(6); end if;

        -- ADD:
        F <= Op_ADD;
        clock <= '0'; wait for 10 ns;
        clock <= '1'; wait for 10 ns;
        t := std_logic_vector(to_unsigned(j+k, 10));
        expected := t(7 downto 0);
        if t /= ('0' & Cout & SUM) then ShowSignals(7); end if;

        -- SET
        F <= Op_SET;
        clock <= '0'; wait for 10 ns;
        clock <= '1'; wait for 10 ns;
        expected := SRI;
        if expected /= ROP2 then ShowSignals(8); end if;

      end loop;
    end loop;

    assert errors = 0
      report integer'image(errors) & " UNEXPECTED RESULTS "
      severity failure;

-- added :
    clock<='X'; -- display the histogram
    wait;
  end process;

end plip;
