#!/bin/bash
# file A3Pv2.3/test3_ALU8/test.sh
# created by Yann Guidon / whygee@f-cpu.org
# version ven. nov. 10 03:25:21 CET 2017
# version lun. nov. 13 02:36:33 CET 2017
# version jeu. nov. 16 03:07:00 CET 2017
# version lun. avril 1 02:58:19 CEST 2019
# version sam. nov. 23 20:22:07 CET 2019 : integration in A3P library

echo
echo  'This script run exhaustive self-tests and will take "a while".'
echo  'You may interrupt it (CTRL+C) if needed. Please report errors.'
echo

if [ ! -f ../proasic3v2/proasic3-obj93.cf  ]; then
  echo -n '--> going to ' ; pushd ../proasic3v2/
  ./build.sh
  echo -n '<-- back to ' ; popd
fi

VER="--std=93"
PA3="-P../proasic3v2"
VHDLIB="../misc/YGREC8_def.vhdl ../misc/YGREC8_utils.vhdl"

run_test () {
# $1 = MSG_name
# $2 = ALU file
# $3 = DEC file
# $4 = flag to compile A3P gates (use $PA3 in argument)
  rm -f alu8 alu8_tb work-obj93.cf *.o &&
  echo -n "ALU $1 standalone: " &&
  ghdl -a $4 $VHDLIB gate_CLA3.vhdl $3 $2 &&
  ghdl -e $4 ALU8 &&
  ./alu8 &&
  echo OK &&
  echo -n "ALU $1 testbench:  " &&
  ghdl -a $4 ALU8_tb.vhdl &&
  ghdl -e $4 ALU8_tb &&
  ./alu8_tb &&
  echo OK
}

# Available decoders :
#../DEC/dec_ALU8_RTL_bool.vhdl  => simple version, ok pour simulation
##../DEC/dec_ALU8_fast_bool.vhdl => minimal logic, basic equations
#../DEC/dec_ALU8_fast_A3P.vhdl  => minimal and fastest logic, mapped to A3P but toggles a lot
##../DEC/dec_ALU8_latch_RTL.vhdl => version with toggle reduction
##../DEC/dec_ALU8_latch_bool.vhdl =>
### Some versions are not included in the A3P test suite.
### Check in the YGREC8 source archives if you need them.

# Available ALU versions:
# ALU8_classic.vhdl => basic version using standard adder
# ALU8_CLA.vhdl     => radix-3 Carry Lookahead Adder
# ALU8_A3P.vhdl     => like above and mapped to A3P tiles

run_test "simple bool"    "ALU8_classic.vhdl" "dec_ALU8_RTL_bool.vhdl"   $PA3 &&
run_test "CLA bool"       "ALU8_CLA.vhdl"     "dec_ALU8_RTL_bool.vhdl"   $PA3 &&
run_test "A3P bool"       "ALU8_A3P.vhdl"     "dec_ALU8_RTL_bool.vhdl"   $PA3 &&
run_test "A3P fast_A3P"   "ALU8_A3P.vhdl"     "dec_ALU8_fast_A3P.vhdl"   $PA3 &&

rm -f alu8 alu8_tb work-obj93.cf *.o &&

echo "OK now let's probe the gates and get the histogram :" &&
ghdl -a $PA3 $VHDLIB gate_CLA3.vhdl dec_ALU8_fast_A3P.vhdl ALU8_A3P.vhdl ALU8_tb.vhdl &&
ghdl -e $PA3 ALU8_tb &&
./alu8_tb -gflip=0 &&

echo 'OK !'
